from unittest import mock

import pytest

import usb.core

from traffic_light.core import ClewareTrafficLight, Color, State, ID_PRODUCT_ORIGINAL
from traffic_light.error import TrafficLightError, MultipleTrafficLightsError


@mock.patch("usb.core.find")
def test_should_initialize_light_by_address(usb_find_mock):
    # when
    light = ClewareTrafficLight(address=10)

    # then
    assert light.address == 10
    assert light.device is not None


@mock.patch("usb.core.find")
def test_should_fail_for_no_connected_light(usb_find_mock):
    # given
    usb_find_mock.return_value = None

    # when & then
    with pytest.raises(TrafficLightError, match="Cleware traffic light not found!"):
        ClewareTrafficLight()


@mock.patch("usb.core.find")
def test_should_fail_for_multiple_lights_whitout_address(usb_find_mock):
    # given
    usb_find_mock.return_value = [mock.MagicMock(), mock.MagicMock()]

    # when & then
    with pytest.raises(
        MultipleTrafficLightsError,
        match="No address is given and there are multiple devices conected! "
        "Use 'print_devices' to see a list of connected devices.",
    ):
        ClewareTrafficLight()


@mock.patch("usb.core.find")
def test_should_fail_for_wrong_color_in_getattr(usb_find_mock):
    # when & then
    with pytest.raises(TrafficLightError, match="Either the given color or state could not be parsed"):
        ClewareTrafficLight().blue_on()


@mock.patch("usb.core.find")
def test_should_fail_for_wrong_state_in_getattr(usb_find_mock):
    # when & then
    with pytest.raises(TrafficLightError, match="Either the given color or state could not be parsed"):
        ClewareTrafficLight().red_foo()


@mock.patch("usb.core.find")
@pytest.mark.parametrize(
    "test_input,expected",
    [
        ((Color.RED, State.ON), (0x10, 0x1)),
        ((Color.RED, State.OFF), (0x10, 0x0)),
        ((Color.YELLOW, State.ON), (0x11, 0x1)),
        ((Color.YELLOW, State.OFF), (0x11, 0x0)),
        ((Color.GREEN, State.ON), (0x12, 0x1)),
        ((Color.GREEN, State.OFF), (0x12, 0x0)),
    ],
)
def test_should_turn_on_led(usb_find_mock, test_input, expected):
    # given
    device_mock = mock.MagicMock(idProduct=ID_PRODUCT_ORIGINAL)
    usb_find_mock.return_value = device_mock
    light = ClewareTrafficLight()

    # when
    light.set_led(test_input[0], test_input[1])

    # then
    device_mock.write.assert_called_once_with(0x02, [0x00, expected[0], expected[1]], timeout=1000)


@mock.patch("usb.core.find")
def test_should_raise_exception_when_light_is_disconneted(usb_find_mock):
    # given
    device_mock = mock.MagicMock()
    device_mock.write.side_effect = usb.core.USBError("[Errno 5] Input/Output Error")
    usb_find_mock.return_value = device_mock
    light = ClewareTrafficLight()

    # when & then
    with pytest.raises(TrafficLightError):
        light.set_led(Color.RED, State.ON)


@mock.patch("usb.core.find")
def test_should_raise_exception_when_color_and_state_invalid(usb_find_mock):
    # given
    device_mock = mock.MagicMock()
    device_mock.write.side_effect = TypeError()
    usb_find_mock.return_value = device_mock
    light = ClewareTrafficLight()

    # when & then
    with pytest.raises(TrafficLightError):
        light.set_led("foo", 1)



import sys
import argparse

from traffic_light.error import TrafficLightError, MultipleTrafficLightsError
from traffic_light.core import ClewareTrafficLight, State, Color, Direction


def main(args=sys.argv[1:]):
    parser = argparse.ArgumentParser(prog="ctl", description="Turns the led of the cleware traffic light on or off")
    parser.add_argument("-r", "--red", choices=["on", "off"], help="Controlls the red led")
    parser.add_argument("-y", "--yellow", choices=["on", "off"], help="Controlls the yellow led")
    parser.add_argument("-g", "--green", choices=["on", "off"], help="Controlls the green led")
    parser.add_argument("-a", "--all", choices=["on", "off"], help="Controlls all leds")
    parser.add_argument(
        "-d",
        "--direction",
        choices=["left", "front", "right", "back", "all"],
        default="all",
        help="Select the direction you want to actuate (requires a trafficlight4)",
    )
    parser.add_argument("-adr", "--address", type=int, help="Specifies which traffic light should be used")
    args = parser.parse_args(args)

    if not vars(args)["all"]:
        all_given_colors = [
            (k, v) for k, v in vars(args).items() if v is not None and k not in ["direction", "address"]
        ]
        if not all_given_colors:
            parser.print_help()
            return 1
    else:
        all_state = vars(args)["all"]
        all_given_colors = [("red", all_state), ("yellow", all_state), ("green", all_state)]

    for color_name, state_name in all_given_colors:
        color = Color[color_name.upper()]
        state = State[state_name.upper()]
        direction = Direction[vars(args)["direction"].upper()]
        address = vars(args)["address"]
        try:
            ClewareTrafficLight(address).set_led(color, state, direction=direction)
        except MultipleTrafficLightsError as exc:
            print(exc, file=sys.stderr)
            ClewareTrafficLight.print_devices()
            return 1
        except TrafficLightError as exc:
            print("Error -> {}".format(exc.message), file=sys.stderr)
            return 1

    return 0


if __name__ == "__main__":
    sys.exit(main())



from enum import IntEnum
from struct import pack

import functools
import usb.core
import usb.util

from traffic_light.error import TrafficLightError, MultipleTrafficLightsError

CTRL_ENDPOINT = 0x02
ID_VENDOR = 0x0D50
ID_PRODUCT_ORIGINAL = 0x0008
ID_PRODUCT_SWITCH = 0x0030
ID_PRODUCTS = (ID_PRODUCT_ORIGINAL, ID_PRODUCT_SWITCH)
INTERFACE = 0


class Color(IntEnum):
    RED = 0x10
    YELLOW = 0x11
    GREEN = 0x12


class State(IntEnum):
    OFF = 0x0
    ON = 0x1


class Direction(IntEnum):
    LEFT = 1
    FRONT = 2
    RIGHT = 4
    BACK = 8
    ALL = LEFT | FRONT | RIGHT | BACK


class ClewareTrafficLight:
    def __init__(self, address=None):
        if address:
            self.address = address
            self.device = self.find_device(address=address)
        elif len(list(ClewareTrafficLight.find_devices())) > 1:
            raise MultipleTrafficLightsError(
                "No address is given and there are multiple devices conected! "
                "Use 'print_devices' to see a list of connected devices."
            )
        else:
            self.device = self.find_device()
        if self.device is None:
            raise TrafficLightError("Cleware traffic light not found!")
        self.reattach = False

    def attach(self):
        """Attaches the device back to the kernel"""
        usb.util.dispose_resources(self.device)
        if self.reattach:
            self.device.attach_kernel_driver(INTERFACE)

    def detach(self):
        """Detaches the device from to kernel so it can be used"""
        if self.device.is_kernel_driver_active(INTERFACE):
            self.device.detach_kernel_driver(INTERFACE)
            self.reattach = True

    @staticmethod
    def find_devices():
        """Returns the raw iterator of all found traffic lights"""
        devices = []

        for product_id in ID_PRODUCTS:
            if devs := usb.core.find(find_all=True, idVendor=ID_VENDOR, idProduct=product_id):
                devices.extend(devs)

        return devices

    @classmethod
    def find_device(self, address=None):
        """Returns a traffic light device, located at the specified address"""

        for product_id in ID_PRODUCTS:
            kwargs = {"idVendor": ID_VENDOR, "idProduct": product_id}
            if address is not None:
                kwargs["address"] = address

            if dev := usb.core.find(**kwargs):
                return dev

    @staticmethod
    def print_devices():
        """Prints a list of all connected traffic lights"""
        devices = ClewareTrafficLight.get_devices()
        for device in devices:
            print(device)

    @staticmethod
    def get_devices():
        """Returns a list of ClewareTrafficLight instances"""
        usb_devices = ClewareTrafficLight.find_devices()
        return [ClewareTrafficLight(d.address) for d in usb_devices]

    def _compute_led_payload(self, color, value, direction):
        if self.device.idProduct == ID_PRODUCT_ORIGINAL:
            return [0x00, color, value]
        elif self.device.idProduct == ID_PRODUCT_SWITCH:
            # The traffic light v4 is acting like a switch. The switches' state
            # are represented as a 16-bit bitfield where bit 0 represents the
            # top-left LED, and all the following bits represent the next LED
            # anticlockwise all the way to bit 11 which represents the state of
            # the back-bottom LED.
            #
            # Updating the bitfield is done by setting every bit we
            # want to update in the mask, then setting the wanted value in the
            # value field. The final result will be equal to:
            #
            #     new_state = (state & ~mask) | value
            #
            # And here is the command one needs to set
            # - action: 8bits <-- Command 11
            # - value: 16 bits (big endian)
            # - mask: 16 bits (big endian)
            #
            color_id = color % 0x10  # Red = 0, Yellow = 1, Green = 2
            color_offset = color_id * 4  # Red = 0-3, Yellow = 4-7, Green = 8-11
            return pack(">BHH", 11, (direction.value << color_offset) * value, direction.value << color_offset)
        else:
            raise TrafficLightError("Unknown product ID")

    def set_led(self, color, value, timeout=1000, direction=Direction.ALL):
        """Sets the given state and color of the attached traffic light

        Attribute:
            color -- the to set color as the enum. E.g. Color.RED
            state -- the state to which it should be set. E.g. State.ON
            address -- the usb address of a specific traffic light
        """
        try:
            self.detach()
            payload = self._compute_led_payload(color, value, direction)
            self.device.write(CTRL_ENDPOINT, payload, timeout=timeout)
        except Exception as exc:
            raise TrafficLightError(str(exc)) from exc
        finally:
            self.attach()

    def __getattr__(self, name):
        """Parses attribut calls in function"""
        args = name.split("_")
        try:
            color = Color[args[0].upper()]
            state = State[args[1].upper()]
        except Exception as exc:
            raise TrafficLightError("Either the given color or state could not be parsed! Exc: {}".format(exc))
        return functools.partial(self.set_led, color, state)

    def __str__(self):
        """Converts instance into string with important imformations"""
        return (
            "== Cleware Traffic Light ==\n"
            "Address: {} \n"
            "IdVendor: {} \n"
            "IdProduct: {}".format(self.address, self.device.idVendor, self.device.idProduct)
        )



from traffic_light.__main__ import main
from traffic_light.core import ID_PRODUCT_ORIGINAL
from unittest import mock
import pytest


@mock.patch("usb.core.find", return_value=mock.MagicMock(idProduct=ID_PRODUCT_ORIGINAL))
@pytest.mark.parametrize(
    "test_input,expected",
    [
        (["--all", "on"], 0),
        (["--all", "off"], 0),
        (["-a", "off"], 0),
    ],
)
def test_set_all_leds(usb_find_mock, test_input, expected):
    assert main(test_input) == expected


@pytest.mark.parametrize("test_input", [(["--all", "foo"]), (["--blue", "on"])])
def test_should_fail_for_wrong_argument_and_value(test_input):
    with pytest.raises(SystemExit):
        main(test_input)


@mock.patch("usb.core.find", return_value=mock.MagicMock(idProduct=ID_PRODUCT_ORIGINAL))
@pytest.mark.parametrize(
    "test_input,expected",
    [
        (["--red", "on"], 0),
        (["--red", "off"], 0),
        (["-r", "off"], 0),
        (["--yellow", "on"], 0),
        (["--yellow", "off"], 0),
        (["-y", "off"], 0),
        (["--green", "on"], 0),
        (["--green", "off"], 0),
        (["-g", "off"], 0),
    ],
)
def test_set_diffrent_color_leds(usb_find_mock, test_input, expected):
    assert main(test_input) == expected


@pytest.mark.parametrize(
    "test_input,expected",
    [
        (["-adr", "2"], 1),
        (["--address", "2"], 1),
    ],
)
def test_should_fail_for_no_given_color(test_input, expected):
    assert main(test_input) == expected


@mock.patch("usb.core.find", return_value=mock.MagicMock(idProduct=ID_PRODUCT_ORIGINAL))
def test_should_fail_for_multiple_lights_whitout_given_address(usb_find_mock):
    # given
    def side_effect(*args, **kwargs):
        if kwargs.get("find_all", False):
            return [mock.MagicMock(), mock.MagicMock()]
        else:
            return mock.MagicMock()

    # usb_find_mock.return_value = [mock.MagicMock(), mock.MagicMock()]
    usb_find_mock.side_effect = side_effect
    # then
    assert main(["--red", "on"]) == 1


@mock.patch("usb.core.find", return_value=mock.MagicMock(idProduct=ID_PRODUCT_ORIGINAL))
def test_should_fail_for_no_connected_light(usb_find_mock):
    # given
    usb_find_mock.return_value = None

    # then
    assert main(["--red", "on"]) == 1



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as desc_file:
    long_description = desc_file.read()

setup(
    name="cleware-traffic-light",
    version="1.0.8",
    description="Python3 modul and cli tool to controll the Cleware traffic light",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Josh Rost",
    author_email="joshua.s.rost@gmail.com",
    url="https://github.com/joshrost/cleware-traffic-light",
    packages=find_packages(),
    install_requires=["pyusb"],
    entry_points={
        "console_scripts": ["ctl=traffic_light.__main__:main"],
    },
)
